import json
import os

import psycopg_pool
from fastapi import FastAPI, Response, HTTPException

import tilekiln
from tilekiln.config import Config
from tilekiln.kiln import Kiln
from tilekiln.tile import Tile
from tilekiln.tileset import Tileset
from tilekiln.storage import Storage

HTTP_TIME = "%a, %d %b %Y %H:%M:%S GMT"

# Constants for MVTs
MVT_MIME_TYPE = "application/vnd.mapbox-vector-tile"

# Constants for environment variable names
# Passing around enviornment variables really is the best way to get this to fastapi
TILEKILN_CONFIG = "TILEKILN_CONFIG"
TILEKILN_URL = "TILEKILN_URL"
TILEKILN_THREADS = "TILEKILN_THREADS"

STANDARD_HEADERS: dict[str, str] = {"Access-Control-Allow-Origin": "*",
                                    "Access-Control-Allow-Methods": "GET, HEAD"}

kiln: Kiln
config: Config
storage: Storage
tilesets: dict[str, Tileset] = {}

# Two types of server are defined - one for static tiles, the other for live generated tiles.
server = FastAPI()
live = FastAPI()

# TODO: Set up middleware for CORS


# TODO: Move elsewhere
def change_tilejson_url(tilejson: str, baseurl: str) -> str:
    modified_tilejson = json.loads(tilejson)
    modified_tilejson["tiles"] = [baseurl + "/{z}/{x}/{y}.mvt"]
    return json.dumps(modified_tilejson)


@server.on_event("startup")
def load_server_config():
    '''Load the config for the server with static pre-rendered tiles'''
    global storage
    global tilesets
    # Because the DB connection variables are passed as standard PG* vars,
    # a plain ConnectionPool() will connect to the right DB
    conn = psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                       check=psycopg_pool.ConnectionPool.check_connection)
    # TODO: Make readonly?

    storage = Storage(conn)
    for tileset in storage.get_tilesets():
        tilesets[tileset.id] = tileset


@live.on_event("startup")
def load_live_config():
    global config
    global storage
    global tilesets
    config = tilekiln.load_config(os.environ[TILEKILN_CONFIG])

    generate_args = {}
    if "GENERATE_PGDATABASE" in os.environ:
        generate_args["dbname"] = os.environ["GENERATE_PGDATABASE"]
    if "GENERATE_PGHOST" in os.environ:
        generate_args["host"] = os.environ["GENERATE_PGHOST"]
    if "GENERATE_PGPORT" in os.environ:
        generate_args["port"] = os.environ["GENERATE_PGPORT"]
    if "GENERATE_PGUSER" in os.environ:
        generate_args["username"] = os.environ["GENERATE_PGUSER"]

    storage_args = {}
    if "STORAGE_PGDATABASE" in os.environ:
        storage_args["dbname"] = os.environ["STORAGE_PGDATABASE"]
    if "STORAGE_PGHOST" in os.environ:
        storage_args["host"] = os.environ["STORAGE_PGHOST"]
    if "STORAGE_PGPORT" in os.environ:
        storage_args["port"] = os.environ["STORAGE_PGPORT"]
    if "STORAGE_PGUSER" in os.environ:
        storage_args["username"] = os.environ["STORAGE_PGUSER"]

    storage_pool = psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                               check=psycopg_pool.ConnectionPool.check_connection,
                                               kwargs=storage_args)

    storage = Storage(storage_pool)

    # Storing the tileset in the dict allows some commonalities in code later
    tilesets[config.id] = Tileset.from_config(storage, config)
    generate_pool = psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                                check=psycopg_pool.ConnectionPool.check_connection,
                                                kwargs=generate_args)
    global kiln
    kiln = Kiln(config, generate_pool)


@server.head("/")
@server.get("/")
@live.head("/")
@live.get("/")
def root():
    raise HTTPException(status_code=404)


@server.head("/favicon.ico")
@server.get("/favicon.ico")
@live.head("/favicon.ico")
@live.get("/favicon.ico")
def favicon():
    return Response("")


@server.head("/{prefix}/tilejson.json")
@server.get("/{prefix}/tilejson.json")
@live.head("/{prefix}/tilejson.json")
@live.get("/{prefix}/tilejson.json")
def tilejson(prefix: str):
    global tilesets
    if prefix not in tilesets:
        raise HTTPException(status_code=404, detail=f'''Tileset {prefix} not found on server.''')
    return Response(content=change_tilejson_url(tilesets[prefix].tilejson,
                                                os.environ[TILEKILN_URL] + f"/{prefix}"),
                    media_type="application/json",
                    headers=STANDARD_HEADERS)


@server.head("/{prefix}/{zoom}/{x}/{y}.mvt")
@server.get("/{prefix}/{zoom}/{x}/{y}.mvt")
def serve_tile(prefix: str, zoom: int, x: int, y: int):
    global tilesets
    if prefix not in tilesets:
        raise HTTPException(status_code=404, detail=f"Tileset {prefix} not found on server.")

    try:
        tile, generated = tilesets[prefix].get_tile(Tile(zoom, x, y))
    except tilekiln.errors.ZoomNotDefined:
        raise HTTPException(status_code=410,
                            detail=f'''Tileset {zoom} not available for tileset {prefix}.''')

    response = b''
    for data in tile.values():
        if data is None:
            raise HTTPException(status_code=404,
                                detail=f"Tile {prefix}/{zoom}/{x}/{y} not found in storage.")
        response += data

    # We use the generated timestamp on the assumption that a specific
    # x/y/z will not be generated twice in the same ms.
    headers: dict[str, str] = {}
    if generated is not None:
        headers = {"Last-Modified": generated.strftime(HTTP_TIME),
                   "E-tag": generated.strftime("%s.%f")}
    return Response(response, media_type=MVT_MIME_TYPE,
                    headers=STANDARD_HEADERS | headers)


@live.head("/{prefix}/{zoom}/{x}/{y}.mvt")
@live.get("/{prefix}/{zoom}/{x}/{y}.mvt")
def live_serve_tile(prefix: str, zoom: int, x: int, y:  int):
    global tilesets
    if prefix not in tilesets:
        raise HTTPException(status_code=404, detail=f"Tileset {prefix} not found on server.")

    # Attempt to serve a stored tile
    try:
        existing, generated = tilesets[prefix].get_tile(Tile(zoom, x, y))
    except tilekiln.errors.ZoomNotDefined:
        raise HTTPException(status_code=410,
                            detail=f'''Tileset {zoom} not available for tileset {prefix}.''')

    response = b''
    missing = []
    for layer, data in existing.items():
        if data is None:
            missing.append(layer)
        else:
            response += data

    # Handle storage hits
    if missing == []:
        headers: dict[str, str] = {}
        if generated is not None:
            headers = {"Last-Modified": generated.strftime(HTTP_TIME),
                       "E-tag": generated.strftime("%s.%f")}
        return Response(response, media_type=MVT_MIME_TYPE,
                        headers=STANDARD_HEADERS | headers)

    # Storage miss, so generate a new tile
    # TODO: partially generate a new tile
    global kiln
    tile = Tile(zoom, x, y)
    new_layers = {layer: kiln.render_layer(layer, tile) for layer in missing}
    # TODO: Make async so tile is saved and response returned in parallel
    generated = tilesets[prefix].save_tile(tile, new_layers)

    mvt = b''.join(new_layers.values()) + b''.join([data for data in existing.values()
                                                    if data is not None])
    if generated is not None:
        headers = {"Last-Modified": generated.strftime(HTTP_TIME),
                   "E-tag": generated.strftime("%s.%f")}
    else:
        headers = {}

    return Response(mvt,
                    media_type=MVT_MIME_TYPE,
                    headers=STANDARD_HEADERS | headers)



import json
import yaml

import fs

from tilekiln.definition import Definition
from tilekiln.errors import ConfigYAMLError, ConfigError
from tilekiln.tile import Tile


class Config:
    def __init__(self, yaml_string: str, filesystem: fs.base.FS):
        '''Create a config from a yaml string
           Creates a config from the yaml string. Any SQL files referenced must be in the
           filesystem.
        '''

        try:
            config = yaml.safe_load(yaml_string)
        except yaml.parser.ParserError:
            raise ConfigYAMLError("Unable to parse config YAML")

        try:
            metadata = config["metadata"]
        except Exception:
            raise ConfigYAMLError("No metadata found in config") from None

        try:
            self.id = metadata["id"]
        except Exception:
            raise ConfigYAMLError("id not found in config metadata") from None
        if not isinstance(self.id, str) or self.id is None:
            raise ConfigYAMLError("metadata.id is not a string") from None

        self.name = metadata.get("name")
        self.description = metadata.get("description")
        self.attribution = metadata.get("attribution")
        self.version = metadata.get("version")
        self.bounds = metadata.get("bounds")
        self.center = metadata.get("center")
        self.__layers = {}
        try:
            for id, l in config.get("vector_layers", {}).items():
                if "\"" in id:
                    raise ConfigError(f"Illegal character \" found in layer name: f{id}")
                if "'" in id:
                    raise ConfigError(f"Illegal character ' found in layer name: f{id}")
                if '\\' in id:
                    raise ConfigError(f"Illegal character \\ found in layer name: f{id}")
                lc = LayerConfig(id, l, filesystem)
                self.__layers[lc.id] = lc

        except Exception:
            raise ConfigError("Unable to process vector_layers")

        if self.__layers:
            self.minzoom = min([layer.minzoom for layer in self.__layers.values()])
            self.maxzoom = max([layer.maxzoom for layer in self.__layers.values()])
        else:
            self.minzoom = None
            self.maxzoom = None

    def tilejson(self, url) -> str:
        '''Returns a TileJSON'''

        result = {"tilejson": "3.0.0",
                  "tiles": [f"{url}/{self.id}" + "/{z}/{x}/{y}.mvt"],
                  "attribution": self.attribution,
                  "bounds": self.bounds,
                  "center": self.center,
                  "description": self.description,
                  "maxzoom": self.maxzoom,
                  "minzoom": self.minzoom,
                  "name": self.name,
                  "scheme": "xyz"}

        vector_layers = [{"id": layer.id,
                          "fields": layer.fields,
                          "description": layer.description,
                          "minzoom": layer.minzoom,
                          "maxzoom": layer.maxzoom} for layer in self.__layers.values()]
        result["vector_layers"] = [{k: v for k, v in layer.items() if v is not None}
                                   for layer in vector_layers]

        return json.dumps({k: v for k, v in result.items() if v is not None},
                          sort_keys=True, indent=4)

    def layer_names(self):
        return [id for id in self.__layers.keys()]

    def layer_query(self, layer: str, tile: Tile) -> str:
        return self.__layers[layer].render_sql(tile)

    def layer_queries(self, tile: Tile) -> dict[str, str | None]:
        '''Returns queries for layers

        For layers defined in the config but not present at this zoom None is returned
        '''
        return {name: layer.render_sql(tile)
                for name, layer in self.__layers.items()}


class LayerConfig:
    def __init__(self, id: str, layer_yaml: dict, filesystem: fs.base.FS):
        '''Create a layer config
           Creates a layer config from the config yaml for a layer. Any SQL files referenced must
           be in the filesystem.
        '''
        self.id = id
        self.description = layer_yaml.get("description")
        self.fields = layer_yaml.get("fields", {})
        self.definitions: list[Definition] = []
        self.geometry_type = set(layer_yaml.get("geometry_type", []))

        self.__definitions = set()
        for definition in layer_yaml.get("sql", []):
            self.__definitions.add(Definition(id, definition, filesystem))

        self.minzoom = min({d.minzoom for d in self.__definitions})
        self.maxzoom = max({d.maxzoom for d in self.__definitions})

    def render_sql(self, tile: Tile) -> str | None:
        '''Returns the SQL for a layer, given a tile, or None if it is outside the zoom range
           of the definitions
        '''
        if tile.zoom > self.maxzoom or tile.zoom < self.minzoom:
            return None

        # Match the first definition for the layer
        for d in self.__definitions:
            if tile.zoom <= d.maxzoom and tile.zoom >= d.minzoom:
                return d.render_sql(tile)

        return None



import jinja2 as j2

import fs

from tilekiln.tile import Tile
from tilekiln.errors import DefinitionError

DEFAULT_EXTENT = 4096
DEFAULT_BUFFER = 0

# Invariants of web mercator
HALF_WORLD = 20037508.34

j2Environment = j2.Environment(loader=j2.BaseLoader(), lstrip_blocks=True, trim_blocks=True)


class Definition:
    def __init__(self, id: str, definition_yaml, filesystem: fs.base.FS):
        self.id = id

        try:
            self.minzoom = definition_yaml["minzoom"]
        except KeyError:
            raise DefinitionError(f"Layer {id} is missing minzoom on a definition") from None
        try:
            self.maxzoom = definition_yaml["maxzoom"]
        except KeyError:
            raise DefinitionError(f"Layer {id} is missing maxzoom on a definition") from None

        self.extent = definition_yaml.get("extent", DEFAULT_EXTENT)
        self.buffer = definition_yaml.get("buffer", DEFAULT_BUFFER)

        # TODO: Let is use directories so one file can include others.
        filename = definition_yaml["file"]
        try:
            self.__template = j2Environment.from_string(filesystem.readtext(filename))
        except fs.errors.ResourceNotFound:
            raise DefinitionError(f"Layer {id} is missing is missing file {filename}") from None

    def render_sql(self, tile: Tile) -> str:
        '''Generate the SQL for a layer
        '''

        # Tile validity constraints. x/y are checked by Tile class
        assert tile.zoom >= self.minzoom
        assert tile.zoom <= self.maxzoom

        # See https://postgis.net/docs/ST_AsMVT.html for SQL source

        inner = self.__template.render(zoom=tile.zoom, x=tile.x, y=tile.y,
                                       bbox=tile.bbox(self.buffer/self.extent),
                                       unbuffered_bbox=tile.bbox(0),
                                       extent=self.extent,
                                       buffer=self.buffer,
                                       tile_length=tile_length(tile),
                                       tile_area=tile_length(tile)**2,
                                       coordinate_length=tile_length(tile)/self.extent,
                                       coordinate_area=(tile_length(tile)/self.extent)**2)

        # TODO: Use proper escaping for self.id in SQL
        return (f'''WITH mvtgeom AS -- {self.id}/{tile.zoom}/{tile.x}/{tile.y}\n(\n''' +
                inner + f'''\n)\nSELECT ST_AsMVT(mvtgeom.*, '{self.id}', {self.extent})\n''' +
                '''FROM mvtgeom;''')


def tile_length(tile) -> float:
    '''Returns the length of a tile, in projected units
    '''
    # -1 for half vs full world
    return HALF_WORLD/(2**(tile.zoom-1))



import datetime
import json
import sys
from collections.abc import Collection, Iterator, Sequence
from typing import Optional

import click
import psycopg.rows
import psycopg_pool
from psycopg import sql

import tilekiln.errors

from tilekiln.metric import Metric
from tilekiln.tile import Tile
from tilekiln.tileset import Tileset

METADATA_TABLE = "metadata"
GENERATE_STATS_TABLE = "generate_stats"
TILE_STATS_TABLE = "tile_stats"

# Lower percentiles are typically not interesting, because generally the
# smallest 50% of tiles are identical water tiles or something similarly
# sparse. Where the data gets interesting is p95 and above.
PERCENTILES = [0.0, 0.25, 0.50, 0.75, 0.90, 0.95, 0.99, 0.999, 1.0]


class Storage:
    '''
    Storage is an object representing a tile storage, backed by a PostgreSQL database

    A Storage contains tiles and metadata about tilesets, and has functions to update
    tiles and metadata based on the ID
    '''
    def __init__(self, pool: psycopg_pool.ConnectionPool, schema: str = "tilekiln"):
        self.__pool = pool
        self.__schema = schema

    '''
    Methods that manipulate schema-related stuff and don't involve any tiles
    '''
    def create_schema(self) -> None:
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                # Perform one-time setup using CREATE ... IF NOT EXISTS
                # This is safe to rerun multiple times
                cur.execute(f'''CREATE SCHEMA IF NOT EXISTS "{self.__schema}"''')
                self.__setup_stats(cur)
                self.__setup_metadata(cur)
            conn.commit()

    '''
    Methods for tilesets
    '''
    def create_tileset(self, id: str, layers: list[str],
                       minzoom: int, maxzoom: int, tilejson: str) -> None:
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                self.__set_metadata(cur, id, layers, minzoom, maxzoom, tilejson)

                self.__setup_tables(cur, id, layers, minzoom, maxzoom)
            conn.commit()

    def remove_tileset(self, id: str) -> None:
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                cur.execute(f'''DELETE FROM "{self.__schema}"."{METADATA_TABLE}" WHERE id = %s''',
                            (id,))
                cur.execute(f'''DROP TABLE "{self.__schema}"."{id}" CASCADE''')
                cur.execute(f'''DELETE FROM "{self.__schema}"."{TILE_STATS_TABLE}" WHERE id = %s''',
                            (id,))
            conn.commit()

    def get_tilesets(self) -> Iterator[Tileset]:
        '''
        Gets all tilesets in the storage
        '''

        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT id, layers, minzoom, maxzoom, tilejson
                                FROM "{self.__schema}"."{METADATA_TABLE}"''')
                for record in cur:
                    yield Tileset(self, record["id"], record["layers"],
                                  record["minzoom"], record["maxzoom"],
                                  json.dumps(record["tilejson"]))

    def get_tileset_ids(self) -> Iterator[str]:
        '''
        Get only the tileset IDs
        '''

        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT id
                                FROM "{self.__schema}"."{METADATA_TABLE}"''')
                for record in cur:
                    yield record["id"]

    def get_tileset(self, id: str) -> Tileset:
        '''
        Fetch a specific tileset
        '''

        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(sql.SQL('''SELECT id, layers, minzoom, maxzoom, tilejson FROM {}.{}''')
                            .format(sql.Identifier(self.__schema), sql.Identifier(METADATA_TABLE)) +
                            sql.SQL('''WHERE id = %s'''), (id, ))
                result = cur.fetchone()
                if result is None:
                    raise tilekiln.errors.TilesetMissing

                return Tileset(self, result["id"], result["layers"],
                               result["minzoom"], result["maxzoom"],
                               json.dumps(result["tilejson"]))

    def get_layer_ids(self, id: str) -> list[str]:
        '''Get layers of a specific tileset
        '''
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(sql.SQL('''SELECT layers FROM {}.{}''')
                            .format(sql.Identifier(self.__schema), sql.Identifier(METADATA_TABLE)) +
                            sql.SQL('''WHERE id = %s'''), (id, ))
                result = cur.fetchone()
                if result is None:
                    raise tilekiln.errors.TilesetMissing
                return result["layers"]

    ''' Methods for metrics'''
    def metrics(self) -> Collection[Metric]:
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT id, zoom, num_tiles, size, percentiles
                                FROM "{self.__schema}"."{TILE_STATS_TABLE}"''')
                return [Metric(**record) for record in cur]

    def update_metrics(self) -> None:
        tilesets = self.get_tilesets()
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                for tileset in tilesets:
                    self.__update_tileset_metrics(cur, tileset)
                conn.commit()

    '''Methods that set/get metadata'''
    def set_metadata(self, id: str, layers: list[str],
                     minzoom: int, maxzoom: int, tilejson: str) -> None:
        '''
        Saves metadata into storage

        This just wraps __set_metadata, which requires a cursor
        '''
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                self.__set_metadata(cur, id, layers, minzoom, maxzoom, tilejson)
            conn.commit()

    # TODO: Should the various get_* functions be separate? The query has to fetch from the
    # DB each time, but only tilejson needs URL. Not an urgent issue.
    def get_tilejson(self, id: str, url: str) -> str:
        '''Gets the tilejson for a layer from storage.'''
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT tilejson
                                FROM "{self.__schema}"."{METADATA_TABLE}"
                                WHERE id = %s''', (id,))
                result = cur.fetchone()
                if result is None:
                    # TODO: raise exception and handle it at the calling level
                    click.echo(f"Failed to retrieve tilejson for id {id}, "
                               f"does it exist in storage DB?",
                               err=True)
                    sys.exit(1)
                tilejson = result["tilejson"]
                tilejson["tiles"] = [f"{url}" + "/{z}/{x}/{y}.mvt"]
                return json.dumps(tilejson)

    # TODO: Get rid of get_minzoom/maxzoom functions and use get_tileset
    def get_minzoom(self, id: str):
        '''Gets the minzoom for a layer from storage.'''
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT minzoom
                                FROM "{self.__schema}"."{METADATA_TABLE}"
                                WHERE id = %s''', (id,))
                result = cur.fetchone()
                if result is None:
                    # TODO: raise exception and handle it at the calling level
                    click.echo(f"Failed to retrieve minzoom for id {id}, "
                               f"does it exist in storage DB?",
                               err=True)
                    sys.exit(1)
                return result["minzoom"]

    def get_maxzoom(self, id):
        '''Gets the minzoom for a layer from storage.'''
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT maxzoom
                                FROM "{self.__schema}"."{METADATA_TABLE}"
                                WHERE id = %s''', (id,))
                result = cur.fetchone()
                if result is None:
                    # TODO: raise exception and handle it at the calling level
                    click.echo(f"Failed to retrieve minzoom for id {id}, "
                               "does it exist in storage DB?", err=True)
                    sys.exit(1)
                return result["maxzoom"]

    '''
    Methods that involve saving, fetching, and deleting tiles
    '''
    def delete_tiles(self, id: str, tiles: set[Tile]):
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                for tile in tiles:
                    self.__delete_tile(cur, id, tile)
            conn.commit()

    def delete_tilelayers(self, id: str, tilelayers: dict[Tile, set[str]]):
        allowed_layers = self.get_layer_ids(id)
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                for tile, layers in tilelayers.items():
                    if layers.difference(allowed_layers) != set():
                        raise tilekiln.errors.LayerNotDefined(
                            f"Layers{layers.difference(allowed_layers)} not defined for {id}")
                    self.__delete_tilelayer(cur, id, tile, layers)
            conn.commit()

    def truncate_tables(self, id: str, zooms: Optional[Sequence[int]] = None):
        if zooms is None:
            zooms = range(self.get_minzoom(id), self.get_maxzoom(id)+1)
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                for zoom in zooms:
                    self.__truncate_table(cur, id, zoom)
            conn.commit()

    def get_tile(self, id: str, tile: Tile) -> tuple[dict[str, bytes | None],
                                                     datetime.datetime | None]:
        tileset = self.get_tileset(id)
        if tile.zoom > tileset.maxzoom or tile.zoom < tileset.minzoom:
            raise tilekiln.errors.ZoomNotDefined
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                query = (sql.SQL('SELECT GREATEST(')
                         + _generated_columns(tileset.layers)
                         + sql.SQL(') AS generated,')
                         + _data_columns(tileset.layers)
                         + sql.SQL('FROM {}.{}').format(sql.Identifier(self.__schema),
                                                        sql.Identifier(id))
                         + sql.SQL('WHERE zoom = %s AND x = %s AND y = %s'))
                cur.execute(query, (tile.zoom, tile.x, tile.y), binary=True)
                result = cur.fetchone()
                if result is None:
                    return {layer: None for layer in tileset.layers}, None
                return {layer: result[f"{layer}_data"]
                        for layer in tileset.layers}, result["generated"]

    def get_tile_details(self, id: str, tile: Tile) -> dict[str,
                                                            tuple[bytes, datetime.datetime] | None]:

        details: dict[str, tuple[bytes, datetime.datetime] | None] = {}

        tileset = self.get_tileset(id)
        if tile.zoom > tileset.maxzoom or tile.zoom < tileset.minzoom:
            raise tilekiln.errors.ZoomNotDefined
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                query = (sql.SQL('SELECT ')
                         + _generated_columns(tileset.layers)
                         + sql.SQL(',')
                         + _data_columns(tileset.layers)
                         + sql.SQL('FROM {}.{}').format(sql.Identifier(self.__schema),
                                                        sql.Identifier(id))
                         + sql.SQL('WHERE zoom = %s AND x = %s AND y = %s'))
                cur.execute(query, (tile.zoom, tile.x, tile.y), binary=True)
                result = cur.fetchone()
                for layer in tileset.layers:
                    if result is not None and result[f"{layer}_data"] is not None:
                        details[layer] = (result[f"{layer}_data"], result[f"{layer}_generated"])
                    else:
                        details[layer] = None
                return details

    def save_tile(self, id: str, tile: Tile,
                  layers: dict[str, bytes], render_time=0) -> datetime.datetime | None:
        tileset = self.get_tileset(id)
        if tile.zoom > tileset.maxzoom or tile.zoom < tileset.minzoom:
            raise tilekiln.errors.ZoomNotDefined
        if set(layers.keys()) - set(tileset.layers):
            raise tilekiln.errors.Error("Rendered tile contains layers not known to storage")
        tablename = f"{id}_z{tile.zoom}"

        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                # TODO: This statement unconditionally writes the row even if it's unchanged. It
                # shouldn't. Adding WHERE tile != EXCLUDED.tile would help, but then it would
                # return zero rows if the contents are the same. The method here instead results
                # in extra writes but does preserve the datetime.

                # The layers used are the incoming tile layers to allow for partial writes

                # Only the layers present in the new tile are
                data_upsert = [sql.SQL("{} = EXCLUDED.{}").format(sql.Identifier(f"{layer}_data"),
                                                                  sql.Identifier(f"{layer}_data"))
                               for layer in layers]
                time_upsert = [sql.SQL("{generated} = CASE WHEN store.{data} IS "
                                       "DISTINCT FROM EXCLUDED.{data} THEN statement_timestamp() "
                                       # It's possible the timestamp is invalidly null,
                                       # so coalesce it with current timestamp
                                       "ELSE COALESCE(store.{generated}, statement_timestamp()) "
                                       "END")
                               .format(generated=sql.Identifier(f"{layer}_generated"),
                                       data=sql.Identifier(f"{layer}_data"))
                               for layer in layers]

                # arguments to the query
                data_idents = [sql.Placeholder(layer) for layer in layers]

                q = (sql.SQL("INSERT INTO {}.{} AS store\n").format(sql.Identifier(self.__schema),
                                                                    sql.Identifier(tablename))
                     + sql.SQL("(zoom, x, y, ") + _data_columns(layers) + sql.SQL(", ")
                     + _generated_columns(layers) + sql.SQL(")\n")
                     + sql.SQL("VALUES ({}, {}, {}, ").format(sql.Literal(tile.zoom),
                                                              sql.Literal(tile.x),
                                                              sql.Literal(tile.y))
                     + sql.SQL(", ").join(data_idents) + sql.SQL(",\n")
                     + sql.SQL(", ").join([sql.SQL("statement_timestamp()") for _ in layers])
                     + sql.SQL(")\n")
                     + sql.SQL("ON CONFLICT (zoom, x, y)\n")
                     + sql.SQL("DO UPDATE SET ")
                     + sql.SQL(", ").join([*data_upsert, *time_upsert])
                     # This has to operate on tileset.layers because we want the greatest date
                     # of *any* layer, not just the layers in the incoming tile
                     + sql.SQL("\nRETURNING GREATEST(") + _generated_columns(tileset.layers)
                     + sql.SQL(") AS generated"))

                cur.execute(q, layers)

                result = cur.fetchone()
                if result is None:
                    return None
                return result["generated"]

    def __setup_metadata(self, cur) -> None:
        ''' Create the metadata table in storage. This is safe to rerun
        '''
        # TODO: Updating metadata table schema??
        # Probably can only be done on a major version upgrade

        cur.execute(f'''CREATE TABLE IF NOT EXISTS "{self.__schema}"."{METADATA_TABLE}" (\n'''
                    '''id text PRIMARY KEY,\n'''
                    '''active boolean NOT NULL DEFAULT TRUE,\n'''
                    '''layers text[],\n'''
                    '''minzoom smallint NOT NULL,\n'''
                    '''maxzoom smallint NOT NULL,\n'''
                    '''tilejson jsonb NOT NULL)''')

    def __set_metadata(self, cur, id: str, layers: list[str], minzoom, maxzoom, tilejson):
        '''
        Sets metadata using a cursor

        This is separate from set_metadata because sometimes it needs
        calling within a transaction
        '''

        query = (sql.SQL("INSERT INTO {}.{}\n").format(sql.Identifier(self.__schema),
                                                       sql.Identifier(METADATA_TABLE))
                 + sql.SQL("(id, minzoom, maxzoom, layers, tilejson)\n")
                 + sql.SQL("VALUES (%s, %s, %s, %s, %s)\n")
                 + sql.SQL("ON CONFLICT (id)\n")
                 + sql.SQL("DO UPDATE SET minzoom = EXCLUDED.minzoom,\n")
                 + sql.SQL("maxzoom = EXCLUDED.maxzoom,\n")
                 + sql.SQL("tilejson = EXCLUDED.tilejson"))
        cur.execute(query, (id, minzoom, maxzoom, layers, tilejson))

    def __setup_stats(self, cur) -> None:
        '''Create the stats tables.

        One table has tile generation stats, the other has tile storage stats.
        '''
        # Because we're just storing counters for prometheus here and unlogged table is fine.
        # Periodic resets are okay.
        # It's necessary to store this in-db since we might call tilerender more than once
        # in a polling interval.
        # TODO: Use this table
        cur.execute(f'''CREATE UNLOGGED TABLE IF NOT EXISTS
            "{self.__schema}"."{GENERATE_STATS_TABLE}" (
            id text,
            zoom smallint,
            num_rendered integer DEFAULT 0,
            time_rendered interval DEFAULT '0',
            PRIMARY KEY (id, zoom)
        )
        ''')

        # This caches information on the number of tiles. Prometheus can be called every 15 seconds
        # and doing a sequential scan that often is a bad idea
        cur.execute(f'''CREATE TABLE IF NOT EXISTS "{self.__schema}"."{TILE_STATS_TABLE}" (
            id text,
            zoom smallint,
            num_tiles integer NOT NULL,
            size bigint NOT NULL,
            percentiles double precision[][] NOT NULL,
            PRIMARY KEY (id, zoom),
            CHECK (array_length(percentiles, 1) = 2)
        )
        ''')

    def __update_tileset_metrics(self, cur, tileset: Tileset) -> None:
        id = tileset.id
        minzoom = tileset.minzoom
        maxzoom = tileset.maxzoom
        for zoom in range(minzoom, maxzoom+1):
            # This SQL statement needs to handle the case of an empty table.
            # Except for COUNT(*) the aggregate functions return NULL for
            # no rows, which is a problem. One option would be to save
            # {{}, {}} as the array but 2-d empty arrays don't really work
            # in PostgreSQL. Instead, we return 0 for all metrics.
            #
            # We set jit to ON as it is faster when the tables are large, but
            # jit is commonly disabled on tile rendering servers because it
            # slows down rendering queries.
            # TODO: Consider if it would be better to completely skip the row
            #       and emit no metric.
            cur.execute('SET LOCAL jit TO ON;')

            length = sql.SQL("+").join([sql.SQL("length({})")
                                        .format(sql.Identifier(f"{layer}_data"))
                                        for layer in tileset.layers])

            query = (sql.SQL("INSERT INTO {}.{}\n").format(sql.Identifier(self.__schema),
                                                           sql.Identifier(TILE_STATS_TABLE))
                     + sql.SQL("SELECT %(id)s AS id, %(zoom)s AS zoom,")
                     + sql.SQL("COUNT(*) AS num_tiles,\n"
                               + "COALESCE (SUM({}), 0) AS size,\n").format(length)

                     + sql.SQL("ARRAY[%(percentile)s, "
                               + "COALESCE(PERCENTILE_CONT(%(percentile)s::double precision[])")
                     + sql.SQL("WITHIN GROUP (ORDER BY {}),\n").format(length)
                     + sql.SQL("array_fill(0, ARRAY[array_length(%(percentile)s, 1)]))] "
                               + "AS percentiles\n").format(length, length)
                     + sql.SQL("FROM {}.{}").format(sql.Identifier(self.__schema),
                                                    sql.Identifier(f"{id}_z{zoom}"))
                     + sql.SQL("ON CONFLICT (id, zoom)\n"
                               + "DO UPDATE SET num_tiles = EXCLUDED.num_tiles,\n"
                               + "size = EXCLUDED.size,\n"
                               + "percentiles = EXCLUDED.percentiles;"))

            cur.execute(query, {'id': id, 'zoom': zoom, 'percentile': PERCENTILES})

    def __setup_tables(self, cur, id: str, layers: list[str],
                       minzoom: int, maxzoom: int) -> None:
        '''Create the tile storage tables

        This creates the tile storage tables. It intentionally
        does not try to overwrite existing tables.
        '''

        columns = [sql.SQL('zoom smallint CHECK (zoom >= {} AND zoom <= {})')
                   .format(sql.Literal(minzoom), sql.Literal(maxzoom)),
                   sql.SQL('x int CHECK (x >= 0 AND x < 1 << zoom)'),
                   sql.SQL('y int CHECK (y >= 0 AND y < 1 << zoom)')]

        columns += [sql.SQL("{} timestamptz")
                    .format(sql.Identifier(f'{layer}_generated'))
                    for layer in layers]
        columns += [sql.SQL("{} bytea").format(sql.Identifier(f'{layer}_data'))
                    for layer in layers]
        columns += [sql.SQL("PRIMARY KEY (zoom, x, y)")]

        query = (sql.SQL('''CREATE TABLE {}.{} (\n''').format(sql.Identifier(self.__schema),
                                                              sql.Identifier(id))
                 + sql.SQL(',\n').join(columns)
                 + sql.SQL(''') PARTITION BY LIST (zoom)'''))

        cur.execute(query)
        for zoom in range(minzoom, maxzoom+1):
            tablename = f"{id}_z{zoom}"
            query = (sql.SQL('''CREATE TABLE {}.{}\n''').format(sql.Identifier(self.__schema),
                                                                sql.Identifier(tablename))
                     + sql.SQL('''PARTITION OF {}.{}\n''').format(sql.Identifier(self.__schema),
                                                                  sql.Identifier(id))
                     + sql.SQL('''FOR VALUES IN ({})''').format(sql.Literal(zoom)))

            cur.execute(query)

    def __truncate_table(self, cur, id: str, zoom: int) -> None:
        '''Remove every tile from a particular tileset and zoom'''
        tablename = f"{id}_z{zoom}"
        cur.execute(f'''TRUNCATE TABLE "{self.__schema}"."{tablename}"''')

    def __delete_tile(self, cur, id: str, tile: Tile):
        '''Delete an individual tile

        How this is implemented is not ideal for long lists of tiles
        to delete, but generally a long list is an entire zoom or a box.

        In the former case it is implemented as __truncate_table, and
        the latter case is not implemented but would take min/max x/y.
        '''
        cur.execute(f'''DELETE FROM "{self.__schema}"."{id}"
                        WHERE zoom = %s AND x = %s AND y = %s''',
                    (tile.zoom, tile.x, tile.y))

    def __delete_tilelayer(self, cur, id: str, tile: Tile, layers: set[str]):
        '''Delete specific layers from a tile
        '''
        tablename = f"{id}_z{tile.zoom}"
        data = sql.SQL(', ').join([sql.SQL("{} = NULL").format(sql.Identifier(f"{layer}_data"))
                                   for layer in layers])
        time = sql.SQL(', ').join([sql.SQL("{} = NULL").format(sql.Identifier(f"{layer}_generated"))
                                   for layer in layers])
        cur.execute(sql.SQL('''UPDATE {}.{}\n''').format(sql.Identifier(self.__schema),
                                                         sql.Identifier(tablename))
                    + sql.SQL('''SET ''') + data + sql.SQL(', ') + time
                    + sql.SQL('''\nWHERE x = %s AND y = %s'''), (tile.x, tile.y))


def _data_columns(layers) -> sql.Composable:
    return sql.SQL(', ').join([sql.Identifier(f"{layer}_data")
                               for layer in layers])


def _generated_columns(layers) -> sql.Composable:
    return sql.SQL(', ').join([sql.Identifier(f"{layer}_generated")
                               for layer in layers])



from dataclasses import dataclass


@dataclass(kw_only=True, frozen=True)
class Metric:
    """ Class for a metric about a tileset in storage """
    id: str
    zoom: int
    num_tiles: int
    size: int
    percentiles: dict[float, float]



from __future__ import annotations
from dataclasses import dataclass
import datetime

import tilekiln.errors
from tilekiln.config import Config
from tilekiln.tile import Tile

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from tilekiln.storage import Storage


@dataclass
class Tileset:
    '''A set of tiles in storage

    A tileset must always have the associated DB entries with tilejson/etc
    TODO: How to handle populating the DB
    '''

    storage: Storage
    id: str
    layers: list[str]
    minzoom: int
    maxzoom: int
    tilejson: str

    @classmethod
    def from_config(cls, storage: Storage, config: Config):
        '''Create a tileset from a Storage and Config'''
        return cls(storage, config.id, config.layer_names(), config.minzoom, config.maxzoom,
                   config.tilejson('REPLACED_BY_SERVER'))

    # todo: this should really be in storage
    @classmethod
    def from_id(cls, storage: Storage, id: str) -> Tileset:
        '''
        Create a tileset from a Storage and id

        This pulls the metadata from the storage
        '''

        layers = [layer for layer in storage.get_layer_ids(id)]
        minzoom = storage.get_minzoom(id)
        maxzoom = storage.get_maxzoom(id)
        tilejson = storage.get_tilejson(id, 'REPLACED_BY_SERVER')
        return cls(storage, id, layers, minzoom, maxzoom, tilejson)

    def prepare_storage(self) -> None:
        self.storage.create_tileset(self.id, self.layers, self.minzoom, self.maxzoom,
                                    self.tilejson)

    def update_storage_metadata(self) -> None:
        '''Sets the metadata in storage'''
        self.storage.set_metadata(self.id, self.layers, self.minzoom, self.maxzoom,
                                  self.tilejson)

    def get_tile(self, tile: Tile) -> tuple[dict[str, bytes | None], datetime.datetime | None]:
        if tile.zoom < self.minzoom or tile.zoom > self.maxzoom:
            raise tilekiln.errors.ZoomNotDefined
        return self.storage.get_tile(self.id, tile)

    def save_tile(self, tile: Tile, layers: dict[str, bytes]) -> datetime.datetime | None:
        if tile.zoom < self.minzoom or tile.zoom > self.maxzoom:
            raise tilekiln.errors.ZoomNotDefined
        return self.storage.save_tile(self.id, tile, layers)



import os

import click
import uvicorn

import tilekiln
import tilekiln.dev
import tilekiln.server


@click.group()
def serve() -> None:
    '''Commands for tile serving.

    All tile serving commands serve tiles and a tilejson over HTTP.
    '''
    pass


@serve.command()
@click.option('--config', required=True, type=click.Path(exists=True, dir_okay=False))
@click.option('--bind-host', default='127.0.0.1', show_default=True,
              help='Bind socket to this host.')
@click.option('--bind-port', default=8000, show_default=True,
              type=click.INT, help='Bind socket to this port.')
@click.option('-n', '--num-threads', default=len(os.sched_getaffinity(0)),
              show_default=True, help='Number of worker processes.')
@click.option('--source-dbname')
@click.option('--source-host', type=click.INT)
@click.option('--source-port')
@click.option('--source-username')
@click.option('--base-url', help='Defaults to http://127.0.0.1:8000' +
              ' or the bind host and port')
@click.option('--id', help='Override YAML config ID')
def dev(config: str, bind_host: str, bind_port: int, num_threads: int,
        source_dbname: str, source_host: str, source_port: int, source_username: str,
        base_url: str, id: str) -> None:
    '''Starts a server for development
    '''
    os.environ[tilekiln.dev.TILEKILN_CONFIG] = config
    os.environ[tilekiln.dev.TILEKILN_ID] = id or tilekiln.load_config(config).id

    if base_url is not None:
        os.environ[tilekiln.dev.TILEKILN_URL] = base_url
    else:
        os.environ[tilekiln.dev.TILEKILN_URL] = (f"http://{bind_host}:{bind_port}")
    if source_dbname is not None:
        os.environ["PGDATABASE"] = source_dbname
    if source_host is not None:
        os.environ["PGHOST"] = source_host
    if source_port is not None:
        os.environ["PGPORT"] = str(source_port)
    if source_username is not None:
        os.environ["PGUSER"] = source_username

    uvicorn.run("tilekiln.dev:dev", host=bind_host, port=bind_port, workers=num_threads)


@serve.command()
@click.option('--config', required=True, type=click.Path(exists=True, dir_okay=False))
@click.option('--bind-host', default='127.0.0.1', show_default=True,
              help='Bind socket to this host. ')
@click.option('--bind-port', default=8000, show_default=True,
              type=click.INT, help='Bind socket to this port.')
@click.option('-n', '--num-threads', default=len(os.sched_getaffinity(0)),
              show_default=True, help='Number of worker processes.')
@click.option('--source-dbname')
@click.option('--source-host')
@click.option('--source-port')
@click.option('--source-username')
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port')
@click.option('--storage-username')
@click.option('--base-url', help='Defaults to http://127.0.0.1:8000' +
              ' or the bind host and port')
def live(config: str, bind_host: str, bind_port: int, num_threads: int,
         source_dbname: str, source_host: str, source_port: int, source_username: str,
         storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
         base_url: str) -> None:
    '''Starts a server for pre-generated tiles from DB'''
    os.environ[tilekiln.server.TILEKILN_CONFIG] = config
    os.environ[tilekiln.server.TILEKILN_THREADS] = str(num_threads)

    if base_url is not None:
        os.environ[tilekiln.dev.TILEKILN_URL] = base_url
    else:
        os.environ[tilekiln.dev.TILEKILN_URL] = (f"http://{bind_host}:{bind_port}")
    if source_dbname is not None:
        os.environ["GENERATE_PGDATABASE"] = source_dbname
    if source_host is not None:
        os.environ["GENERATE_PGHOST"] = source_host
    if source_port is not None:
        os.environ["GENERATE_PGPORT"] = str(source_port)
    if source_username is not None:
        os.environ["GENERATE_PGUSER"] = source_username

    if storage_dbname is not None:
        os.environ["STORAGE_PGDATABASE"] = storage_dbname
    if storage_host is not None:
        os.environ["STORAGE_PGHOST"] = storage_host
    if storage_port is not None:
        os.environ["STORAGE_PGPORT"] = str(storage_port)
    if storage_username is not None:
        os.environ["STORAGE_PGUSER"] = storage_username

    uvicorn.run("tilekiln.server:live", host=bind_host, port=bind_port, workers=num_threads)


@serve.command()
@click.option('--bind-host', default='127.0.0.1', show_default=True,
              help='Bind socket to this host. ')
@click.option('--bind-port', default=8000, show_default=True,
              type=click.INT, help='Bind socket to this port.')
@click.option('-n', '--num-threads', default=len(os.sched_getaffinity(0)),
              type=click.INT, show_default=True, help='Number of worker processes.')
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port')
@click.option('--storage-username')
@click.option('--base-url', help='Defaults to http://127.0.0.1:8000' +
              ' or the bind host and port')
def static(bind_host: str, bind_port: int, num_threads: int,
           storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
           base_url: str) -> None:
    '''Starts a server for pre-generated tiles from DB'''

    os.environ[tilekiln.server.TILEKILN_THREADS] = str(num_threads)

    if base_url is not None:
        os.environ[tilekiln.dev.TILEKILN_URL] = base_url
    else:
        os.environ[tilekiln.dev.TILEKILN_URL] = (f"http://{bind_host}:{bind_port}")
    if storage_dbname is not None:
        os.environ["PGDATABASE"] = storage_dbname
    if storage_host is not None:
        os.environ["PGHOST"] = storage_host
    if storage_port is not None:
        os.environ["PGPORT"] = str(storage_port)
    if storage_username is not None:
        os.environ["PGUSER"] = storage_username

    uvicorn.run("tilekiln.server:server", host=bind_host, port=bind_port, workers=num_threads)



import os
import sys

import click
from tqdm import tqdm

import tilekiln

from tilekiln.tile import Tile, layer_frominput
from tilekiln.tilerange import Tilerange
import tilekiln.generator


@click.group()
def generate() -> None:
    '''Commands for tile generation.

    All tile generation commands run queries against the source database which has the
    geospatial data
    '''
    pass


@generate.command()
@click.option('--config', required=True, type=click.Path(exists=True, dir_okay=False))
@click.option('-n', '--num-threads', default=len(os.sched_getaffinity(0)),
              show_default=True, help='Number of worker processes.')
@click.option('--source-dbname')
@click.option('--source-host')
@click.option('--source-port')
@click.option('--source-username')
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port')
@click.option('--storage-username')
@click.option('--progress/--no-progress', help='Display progress bar')
def tiles(config: int, num_threads: int,
          source_dbname: str, source_host: str, source_port: int, source_username: str,
          storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
          progress: bool) -> None:
    '''Generate specific tiles.

    A list of z/x/y tiles is read from stdin and those tiles are generated and saved
    to storage.
    '''

    c = tilekiln.load_config(config)

    tiles = {Tile.from_string(t) for t in sys.stdin}
    threads = min(num_threads, len(tiles))  # No point in more threads than tiles

    click.echo(f"Rendering {len(tiles)} tiles over {threads} threads")

    source_kwargs = {"dbname": source_dbname,
                     "host": source_host,
                     "port": source_port,
                     "user": source_username}
    storage_kwargs = {"dbname": storage_dbname,
                      "host": storage_host,
                      "port": storage_port,
                      "user": storage_username}
    if progress:
        tilekiln.generator.generate(c, source_kwargs, storage_kwargs, tqdm(tiles), threads)
    else:
        tilekiln.generator.generate(c, source_kwargs, storage_kwargs, tiles, threads)


@generate.command()
@click.option('--config', required=True, type=click.Path(exists=True, dir_okay=False))
@click.option('-n', '--num-threads', default=len(os.sched_getaffinity(0)),
              show_default=True, help='Number of worker processes.')
@click.option('--source-dbname')
@click.option('--source-host')
@click.option('--source-port')
@click.option('--source-username')
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port')
@click.option('--storage-username')
@click.option('--min-zoom', type=click.INT, required=True)
@click.option('--max-zoom', type=click.INT, required=True)
@click.option('--progress/--no-progress', help='Display progress bar')
@click.option('--layer', multiple=True)
def zooms(config: int, num_threads: int,
          source_dbname: str, source_host: str, source_port: int, source_username: str,
          storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
          min_zoom: int, max_zoom: int, progress: bool, layer: tuple[str]) -> None:

    if layer == ():
        layers = None
    else:
        layers = set(layer)
    c = tilekiln.load_config(config)

    tiles = Tilerange(min_zoom, max_zoom)
    threads = min(num_threads, len(tiles))  # No point in more threads than tiles
    click.echo(f"Rendering {len(tiles)} tiles over {threads} threads")
    source_kwargs = {"dbname": source_dbname,
                     "host": source_host,
                     "port": source_port,
                     "user": source_username}
    storage_kwargs = {"dbname": storage_dbname,
                      "host": storage_host,
                      "port": storage_port,
                      "user": storage_username}
    if progress:
        tilekiln.generator.generate(c, source_kwargs, storage_kwargs, tqdm(tiles), threads, layers)
    else:
        tilekiln.generator.generate(c, source_kwargs, storage_kwargs, tiles, threads, layers)


@generate.command()
@click.option('--config', required=True, type=click.Path(exists=True, dir_okay=False))
@click.option('-n', '--num-threads', default=len(os.sched_getaffinity(0)),
              show_default=True, help='Number of worker processes.')
@click.option('--source-dbname')
@click.option('--source-host')
@click.option('--source-port')
@click.option('--source-username')
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port')
@click.option('--storage-username')
@click.option('--progress/--no-progress', help='Display progress bar')
def layers(config: int, num_threads: int,
           source_dbname: str, source_host: str, source_port: int, source_username: str,
           storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
           progress: bool) -> None:
    '''Generate specific tile layers.

    A list of z/x/y,layer layers is read from stdin and those are generated and saved
    to storage.
    '''

    c = tilekiln.load_config(config)

    layers = layer_frominput(sys.stdin.read())
    threads = min(num_threads, len(layers))  # No point in more threads than tiles

    click.echo(f"Rendering {len(layers)} tiles over {threads} threads")

    source_kwargs = {"dbname": source_dbname,
                     "host": source_host,
                     "port": source_port,
                     "user": source_username}
    storage_kwargs = {"dbname": storage_dbname,
                      "host": storage_host,
                      "port": storage_port,
                      "user": storage_username}
    if progress:
        tilekiln.generator.generate_tilelayers(c, source_kwargs, storage_kwargs,
                                               tqdm(layers.items()), threads)
    else:
        tilekiln.generator.generate_tilelayers(c, source_kwargs, storage_kwargs,
                                               layers.items(), threads)



import datetime
import sys

import click
import psycopg_pool

import tilekiln

from tilekiln.tile import Tile, layer_frominput
from tilekiln.tileset import Tileset
from tilekiln.storage import Storage


@click.group()
def storage() -> None:
    '''Commands working with tile storage.

    These commands allow creation and manipulation of the tile storage database.
    '''
    pass


@storage.command()
@click.option('--config', required=True, type=click.Path(exists=True, dir_okay=False))
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port', type=click.INT)
@click.option('--storage-username')
@click.option('--id', help='Override YAML config ID')
def init(config: str,
         storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
         id: str) -> None:
    '''Initialize storage for tiles.

    Creates the storage for a tile layer and stores its metadata in the database.
    If the metadata tables have not yet been created they will also be setup.
    '''

    c = tilekiln.load_config(config)

    with psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                     check=psycopg_pool.ConnectionPool.check_connection,
                                     kwargs={"dbname": storage_dbname, "host": storage_host,
                                             "port": storage_port, "user": storage_username
                                             }) as pool:
        storage = Storage(pool)
        storage.create_schema()
        tileset = Tileset.from_config(storage, c)
        tileset.prepare_storage()


@storage.command()
@click.option('--config', type=click.Path(exists=True, dir_okay=False))
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port', type=click.INT)
@click.option('--storage-username')
@click.option('--id', help='Override YAML config ID')
def destroy(config: str,
            storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
            id: str) -> None:
    ''' Destroy storage for tiles'''
    if config is None and id is None:
        raise click.UsageError('''Missing one of '--id' or '--config' options''')

    # No id specified, so load the config for one. We know from above config is not none.
    c = None
    if id is None:
        c = tilekiln.load_config(config)
        id = c.id

    with psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                     check=psycopg_pool.ConnectionPool.check_connection,
                                     kwargs={"dbname": storage_dbname, "host": storage_host,
                                             "port": storage_port, "user": storage_username
                                             }) as pool:
        storage = Storage(pool)
        storage.remove_tileset(id)


@storage.command()
@click.option('--config', type=click.Path(exists=True, dir_okay=False))
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port', type=click.INT)
@click.option('--storage-username')
@click.option('--id', help='Override YAML config ID')
@click.option('-z', '--zoom', required=True, type=click.INT)
@click.option('-x', required=True, type=click.INT)
@click.option('-y', required=True, type=click.INT)
def inspect(config: str,
            storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
            id: str, zoom: int, x: int, y: int) -> None:
    ''' Inspect a tile in storage'''
    if config is None and id is None:
        raise click.UsageError('''Missing one of '--id' or '--config' options''')

    # No id specified, so load the config for one. We know from above config is not none.
    c = None
    if id is None:
        c = tilekiln.load_config(config)
        id = c.id

    with psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                     check=psycopg_pool.ConnectionPool.check_connection,
                                     kwargs={"dbname": storage_dbname, "host": storage_host,
                                             "port": storage_port, "user": storage_username
                                             }) as conn:
        storage = Storage(conn)
        tile = Tile(zoom, x, y)
        mvt = storage.get_tile_details(id, tile)
        size = 0
        click.echo(f"Tile {zoom}/{x}/{y} in {id}")
        for layer, data in mvt.items():
            click.echo(f"   {layer}: {data_info(data)}")
            if data is not None:
                size += len(data[0])
        click.echo(f"Total stored: {size}b")


def data_info(data: tuple[bytes, datetime.datetime] | None):
    if data is None:
        return "undefined"
    if data[1] is None:
        return f"{len(data[0])}b timestamp invalidly null"
    return f"{len(data[0])}b {data[1].isoformat()}"


@storage.command()
@click.option('--config', type=click.Path(exists=True, dir_okay=False))
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port')
@click.option('--storage-username')
@click.option('-z', '--zoom', type=click.INT, multiple=True)
@click.option('--id', help='Override YAML config ID')
def delete(config: str,
           storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
           zoom: tuple[int], id: str) -> None:
    '''Mass-delete tiles from a tileset

    Deletes tiles from a tileset, by zoom, or delete all zooms.
    '''
    if config is None and id is None:
        raise click.UsageError('''Missing one of '--id' or '--config' options''')

    # No id specified, so load the config for one. We know from above config is not none.
    c = None
    if id is None:
        c = tilekiln.load_config(config)
        id = c.id

    with psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                     check=psycopg_pool.ConnectionPool.check_connection,
                                     kwargs={"dbname": storage_dbname, "host": storage_host,
                                             "port": storage_port, "user": storage_username
                                             }) as conn:
        storage = Storage(conn)

        if (len(zoom) == 0):
            storage.truncate_tables(id)
        else:
            storage.truncate_tables(id, zoom)


@storage.command()
@click.option('--config', type=click.Path(exists=True, dir_okay=False))
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port', type=click.INT)
@click.option('--storage-username')
@click.option('--id', help='Override YAML config ID')
def tiledelete(config: str,
               storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
               id: str) -> None:
    '''Delete specific tiles.

    A list of z/x/y tiles is read from stdin and those tiles are deleted from
    storage. The entire list is read before deletion starts.
    '''
    if config is None and id is None:
        raise click.UsageError('''Missing one of '--id' or '--config' options''')

    # No id specified, so load the config for one. We know from above config is not none.
    c = None
    if id is None:
        c = tilekiln.load_config(config)
        id = c.id

    with psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                     check=psycopg_pool.ConnectionPool.check_connection,
                                     kwargs={"dbname": storage_dbname, "host": storage_host,
                                             "port": storage_port, "user": storage_username
                                             }) as pool:
        storage = Storage(pool)

        # TODO: This requires reading all of stdin before starting. This lets it display
        # how many tiles to delete but also means it has to read them all in before starting
        tiles = {Tile.from_string(t) for t in sys.stdin}
        click.echo(f"Deleting {len(tiles)} tiles")
        storage.delete_tiles(id, tiles)


@storage.command()
@click.option('--config', type=click.Path(exists=True, dir_okay=False))
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port', type=click.INT)
@click.option('--storage-username')
@click.option('--id', help='Override YAML config ID')
def layerdelete(config: str,
                storage_dbname: str, storage_host: str, storage_port: int, storage_username: str,
                id: str) -> None:
    '''Delete specific tiles.

    A list of z/x/y tiles is read from stdin and those tiles are deleted from
    storage. The entire list is read before deletion starts.
    '''
    if config is None and id is None:
        raise click.UsageError('''Missing one of '--id' or '--config' options''')

    # No id specified, so load the config for one. We know from above config is not none.
    c = None
    if id is None:
        c = tilekiln.load_config(config)
        id = c.id

    with psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                     check=psycopg_pool.ConnectionPool.check_connection,
                                     kwargs={"dbname": storage_dbname, "host": storage_host,
                                             "port": storage_port, "user": storage_username
                                             }) as pool:
        storage = Storage(pool)

        tilelayers = layer_frominput(sys.stdin.read())
        click.echo(f"Deleting {len(tilelayers)} tiles")
        storage.delete_tilelayers(id, tilelayers)



'''
The code here pulls creates multiple kilns to generate the tiles in parallel
'''
import multiprocessing as mp
from collections.abc import Collection
from typing import Optional

import psycopg_pool

from tilekiln.config import Config
from tilekiln.kiln import Kiln
from tilekiln.storage import Storage
from tilekiln.tile import Tile
from tilekiln.tileset import Tileset


kiln: Kiln
tileset: Tileset
tile_layers: Optional[set[str]]


def setup(config: Config, layers: Optional[set[str]],
          source_kwargs, storage_kwargs) -> None:  # type: ignore[no-untyped-def]
    '''
    Sets up the kiln and tileset for the worker function.
    '''
    global kiln, tileset, tile_layers
    source_pool = psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                              check=psycopg_pool.ConnectionPool.check_connection,
                                              kwargs=source_kwargs)
    kiln = Kiln(config, source_pool)

    storage_pool = psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                               check=psycopg_pool.ConnectionPool.check_connection,
                                               kwargs=storage_kwargs)
    storage = Storage(storage_pool)
    tileset = Tileset.from_config(storage, config)
    tile_layers = layers


def worker(tile: Tile) -> None:
    global kiln, tileset, tile_layers
    # Reuse the tilelayer worker if we're only rendering specific
    if tile_layers is not None:
        tilelayer_worker((tile, tile_layers))
    else:
        try:
            mvt = kiln.render_all(tile)
            tileset.save_tile(tile, mvt)
        except Exception as e:
            print(f"Error generating {tile}")
            raise RuntimeError(f"Error generating {tile}") from e


def tilelayer_worker(work: tuple[Tile, set[str]]) -> None:
    global kiln, tileset
    tile, layers = work
    try:
        new_mvts = {layer: kiln.render_layer(layer, tile) for layer in layers}
        tileset.save_tile(tile, new_mvts)
    except Exception as e:
        print(f"Error generating {tile}")
        raise RuntimeError(f"Error generating {tile}") from e


def generate(config: Config, source_kwargs, storage_kwargs,  # type: ignore[no-untyped-def]
             tiles: Collection[Tile], num_processes: int,
             layers: Optional[set[str]] = None) -> None:

    # If there are no processes and no tiles then there's nothing to do.
    if num_processes == 0 and len(tiles) == 0:
        return

    with mp.Pool(num_processes, setup, (config, layers, source_kwargs, storage_kwargs)) as pool:
        imap_it = pool.imap_unordered(worker, tiles, 100)
        pool.close()
        pool.join()

        # Check for exceptions
        for x in imap_it:
            pass


def generate_tilelayers(config: Config, source_kwargs,
                        storage_kwargs,  # type: ignore[no-untyped-def]
                        tilelayers: Collection[tuple[Tile, set[str]]], num_processes: int) -> None:

    # If there are no processes and no tiles then there's nothing to do.
    if num_processes == 0 and len(tilelayers) == 0:
        return

    with mp.Pool(num_processes, setup, (config, source_kwargs, storage_kwargs)) as pool:
        imap_it = pool.imap_unordered(tilelayer_worker, tilelayers, 100)
        pool.close()
        pool.join()

        # Check for exceptions
        for x in imap_it:
            pass



import os

import fs.osfs

import tilekiln.config


# TODO: Put somewhere else
def load_config(path) -> tilekiln.config.Config:
    '''Loads a config from the filesystem, given a path'''

    full_path = os.path.join(os.getcwd(), path)
    root_path = os.path.dirname(full_path)
    config_path = os.path.relpath(full_path, root_path)
    filesystem = fs.osfs.OSFS(root_path)

    return tilekiln.config.Config(filesystem.open(config_path).read(),
                                  filesystem)
